###################################################
#   ESDOM version 1.1.7                           #
#   http://www5b.biglobe.ne.jp/~hokko2nd/         #
###################################################
package XML::ESDOM_1_1;
use strict;
use XML::ESDOM_1_1::Document;
use XML::ESDOM_1_1::Node;
use XML::ESDOM_1_1::Element;
use XML::ESDOM_1_1::Attr;
use XML::ESDOM_1_1::Text;
use XML::ESDOM_1_1::CDATASection;
use XML::ESDOM_1_1::ProcessingInstruction;
use XML::ESDOM_1_1::DocumentType;
use utf8;
require Exporter;
our %entityTable;
our $errorMessage;
our ( $VERSION, @ISA, @EXPORT );
our %bufferInfo;
our ( $_BI_CURRENT, $_BI_LENGTH );

BEGIN
{
	$VERSION = "1.001007";
	@ISA = qw(Exporter);
	@EXPORT = qw(
		UNKNOWN_NODE
		ELEMENT_NODE
		ATTRIBUTE_NODE
		TEXT_NODE
		CDATA_SECTION_NODE
		ENTITY_REFERENCE_NODE
		ENTITY_NODE
		PROCESSING_INSTRUCTION_NODE
		COMMENT_NODE
		DOCUMENT_NODE
		DOCUMENT_TYPE_NODE
		DOCUMENT_FRAGMENT_NODE
		NOTATION_NODE
		getLastError
		clearLastError
		createDocument
		parse
		parseFromBuffer
		serialize
		serializeToBuffer
		UTF_8
		UTF_16BE
		UTF_16LE
		UTF_16BE_NONBOM
		UTF_16LE_NONBOM
		);


	$entityTable{'amp'} = '&';
	$entityTable{'gt'} = '>';
	$entityTable{'lt'} = '<';
	$entityTable{'apos'} = '\'';
	$entityTable{'quot'} = '"';

	$_BI_CURRENT = 0;
	$_BI_LENGTH = 1;

}

sub TYPE_FILE 	(){1;}
sub TYPE_BUFFER	(){2;}

sub UTF_8			(){1;}
sub UTF_16BE		(){2;}
sub UTF_16LE		(){3;}
sub UTF_16BE_NONBOM	(){4;}
sub UTF_16LE_NONBOM	(){5;}

sub getLastError {
	$errorMessage;
}
sub clearLastError {
	$errorMessage = '';
}

sub createDocument {
	my ( $version, $rootTagName, $doctype ) = @_;
	my $document = new XML::ESDOM_1_1::Document;

	if( $version != 1 && $version ne '' ) {
		$document->_version = $version;
	}
	if( $rootTagName eq '' ) {
		print STDERR "invalid args - createDocument( version, rootTagName, doctype )\n";
		return;
	}
	$document->_setDocumentElement( new XML::ESDOM_1_1::Element( $rootTagName ) );

	if( $doctype ne '' ) {
		$document->setDoctype( $doctype );
	}
	$document->_valid = 1;
	$document;
}

sub parse {
	my $HANDLE = shift;
	binmode( $HANDLE );
	&_parse( \$HANDLE, TYPE_FILE );
}

sub parseFromBuffer {
	my $xmldoc = shift;
	my @info = ( 0, length($xmldoc) );
	$bufferInfo{\$xmldoc} = \@info;
	&_parse( \$xmldoc, TYPE_BUFFER );
}

sub serialize {
	my $document = shift;
	my $encoding = shift;
	my $HANDLE = shift;
	binmode( $HANDLE );
	&_serialize($document, $encoding, $HANDLE, TYPE_FILE);
}

sub serializeToBuffer {
	my $document = shift;
	my $encoding = shift;

	my $output;
	my $result = &_serialize($document, $encoding, \$output, TYPE_BUFFER);

	if( $result ) {
		return $output;
	}
	'';
}

###############################################################################
#

sub _parse {
	my ( $src, $srcType ) = @_;
	my $encoding;
	my $document = new XML::ESDOM_1_1::Document;

	eval {
		my @sample;
		my $buffer;
		my $ch;

		@sample = &_read( $src, $srcType, 2 );
		if( $sample[0] == 0xfe && $sample[1] == 0xff )
		{ # UTF-16 BigEndian
			$encoding = UTF_16BE;
			$buffer = &_readChars( $src, $srcType, $encoding, 5 );
			if( $buffer ne "<?xml" )
			{
				$errorMessage = "document is not XML";
				return 0;
			}
		}
		elsif( $sample[0] == 0xff && $sample[1] == 0xfe )
		{ # UTF-16 LittleEndian
			$encoding = UTF_16LE;
			$buffer = &_readChars( $src, $srcType, $encoding, 5 );
			if( $buffer ne "<?xml" )
			{
				$errorMessage = "document is not XML";
				return 0;
			}
		}
		elsif( $sample[0] == 0x00 && $sample[1] == 0x3c )
		{ # UTF-16 BigEndian without BOM
			$encoding = UTF_16BE;
			$buffer = &_readChars( $src, $srcType, $encoding, 4 );
			if( $buffer ne "?xml" )
			{
				$errorMessage = "document is not XML";
				return 0;
			}
		}
		elsif( $sample[0] == 0x3c && $sample[1] == 0x00 )
		{ # UTF-16 LittleEndian without BOM
			$encoding = UTF_16LE;
			$buffer = &_readChars( $src, $srcType, $encoding, 4 );
			if( $buffer ne "?xml" )
			{
				$errorMessage = "document is not XML";
				return 0;
			}
		}
		elsif( $sample[0] == 0xef && $sample[1] == 0xbb )
		{ # UTF-8 with BOM ?
			&_read( $src, $srcType, \@sample, 1 );
			if( $sample[0] != 0xbf )
			{
				$errorMessage = "unknown BOM";
				return 0;
			}
			$encoding = UTF_8;
			$buffer = &_readChars( $src, $srcType, $encoding, 5 );
			if( $buffer ne "<?xml" )
			{
				$errorMessage = "document is not XML";
				return 0;
			}
		}
		elsif( $sample[0] == 0x3c && $sample[1] == 0x3f )
		{ # UTF-8 without BOM
			$encoding = UTF_8;
			$buffer = &_readChars( $src, $srcType, $encoding, 3 );
			if( $buffer ne "xml" )
			{
				$errorMessage = "document is not XML";
				return 0;
			}
		}
		else
		{
			$errorMessage = "unknown BOM";
			return 0;
		}

		$buffer = &_readChars( $src, $srcType, $encoding, 1 );
		unless( &_isSpace($buffer) )
		{
			$errorMessage = "document is not XML";
			return 0;
		}

		# XMLDecl
		my $dummyNode = new XML::ESDOM_1_1::Element("dummy");
		while(1) {
			$ch = &_skipSpace( $src, $srcType, $encoding );
			last if( $ch eq '?' );
			&_parseAttribute( $src, $srcType, $encoding, $dummyNode, $ch );
		}
		$document->_version = $dummyNode->getAttribute("version");
		if( $encoding == UTF_8 )
		{
			my $enc = $dummyNode->getAttribute("encoding");
			if( $enc !~ /^UTF-8$/i )
			{
				$errorMessage = "unknown encoding type";
				return 0;
			}
		}
		if( $dummyNode->getAttribute("standalone") eq "yes" )
		{
			$document->setStandalone(1);
		}
		elsif( $dummyNode->getAttribute("standalone") eq "no" )
		{
			$document->setStandalone(0);
		}
		$ch = &_readChars( $src, $srcType, $encoding, 1 );
		if( $ch ne '>' )
		{
			$errorMessage = "document is not XML $ch";
			return 0;
		}

		while(1)
		{
			eval{
				$ch = &_skipSpace( $src, $srcType, $encoding );
			};
			if($@){
				if( $document->documentElement )
				{
					$document->_valid = 1;
					return $document;
				}
				return 0;
			}
			if( $ch ne '<' )
			{
				$errorMessage = "document is ungrammatical : ";
				$errorMessage .= &_getCurrentPos( $src, $srcType );
				return 0;
			}
			$ch = &_readChars( $src, $srcType, $encoding, 1 );
			if( $ch eq '?' )
			{ # Processing instructions
				$document->appendChild( &_parseProcessingInstruction($src, $srcType, $encoding) );
			}
			elsif( $ch eq '!' )
			{
				$buffer = &_readChars( $src, $srcType, $encoding, 2 );
				if( $buffer eq '--' )
				{ # Comment
					$document->appendChild( &_parseComment($src, $srcType, $encoding) );
				}
				elsif( $buffer eq 'DO' )
				{
					$buffer = &_readChars( $src, $srcType, $encoding, 5 );
					$ch = &_readChars( $src, $srcType, $encoding, 1 );
					if( $buffer eq 'CTYPE' && &_isSpace($ch) )
					{ # DocumentType
						if( $document->doctype )
						{
							$errorMessage = "document type already exist : ";
							$errorMessage .= &_getCurrentPos( $src,$srcType );
							return 0;
						}
						my $doctype;
						$ch = &_skipSpace( $src, $srcType, $encoding );
						while(1)
						{
							if( $ch eq '>' ) {
								last;
							} elsif( $ch eq '[' )
							{
								$doctype .= $ch;
								while(1)
								{
									$ch = &_readChars( $src, $srcType, $encoding, 1 );
									last if( $ch eq ']' );
									$doctype .= $ch;
								}
							}
							$doctype .= $ch;
							$ch = &_readChars( $src, $srcType, $encoding, 1 );
						}
						$document->setDoctype($doctype);
					}
					else
					{
						$errorMessage = "unknown node : ";
						$errorMessage .= &_getCurrentPos( $src,$srcType );
						return 0;
					}
				}
				else
				{
					$errorMessage = "unknown node : ";
					$errorMessage .= &_getCurrentPos( $src,$srcType );
					return 0;
				}
			}
			elsif( &_isLetter($ch) || $ch eq '_' || $ch eq ':' )
			{
				if( $document->documentElement )
				{
					$errorMessage = "document element is already exist : ";
					$errorMessage .= &_getCurrentPos( $src,$srcType );
					return 0;
				}
				$document->_setDocumentElement( &_parseElement( $src, $srcType, $encoding, $ch ) );
			}
			else
			{
				$errorMessage = "unknown node : ";
				$errorMessage .= &_getCurrentPos( $src,$srcType );
				return 0;
			}
		}
	};
	if($@) {
		print STDERR $@;
		return 0;
	}
	if( $document && $document->_valid ) {
		return $document;
	}
	return 0;
}

sub _parseProcessingInstruction {
	my ( $src, $srcType, $encoding ) = @_;

	my $ch;
	my $target;
	my $fFirst = 1;
	while(1)
	{
		$ch = &_readChars( $src, $srcType, $encoding, 1 );
		last if( &_isSpace($ch) || $ch eq '?' );
		if( $fFirst )
		{
			$fFirst = 0;
			unless( &_isLetter($ch) && $ch ne '_' && $ch ne ':' )
			{
				$errorMessage = "Invalid PI target : ";
				$errorMessage .= &_getCurrentPos( $src, $srcType );
				die $errorMessage;
			}
		}
		else
		{
			if( !&_isLetter($ch) && !&_isDigit($ch)
				 && $ch ne '.' && $ch ne '-' && $ch ne '_' && $ch ne ':'
				&& !&_isCombiningChar($ch) && !&_isExtender($ch) )
			{
				$errorMessage = "Invalid PI target : ";
				$errorMessage .= &_getCurrentPos( $src, $srcType );
				die $errorMessage;
			}
		}
		$target .= $ch;
	}

	if( $target ne "xml" )
	{
		$errorMessage = "invalid PI : ";
		$errorMessage .= &_getCurrentPos( $src, $srcType );
		die $errorMessage;
	}

	my $newPI = new XML::ESDOM_1_1::ProcessingInstruction($target);

	if( $ch eq '?' )
	{
		$ch = &_readChars( $src, $srcType, $encoding, 1 );
		if( $ch ne '>' )
		{
			$errorMessage = "invalid PI : ";
			$errorMessage .= &_getCurrentPos( $src, $srcType );
			die $errorMessage;
		}
	}
	else
	{
		my $value;
		$ch = &_skipSpace( $src, $srcType, $encoding );
		while(1)
		{
			if( $ch eq '?' )
			{
				$ch = &_readChars( $src, $srcType, $encoding, 1 );
				last if( $ch eq '>' );
				$value .= '?';
			}
			if( !&_isChar($ch) )
			{
				$errorMessage = "Invalid PI value : ";
				$errorMessage .= &_getCurrentPos( $src, $srcType );
				die $errorMessage;
			}
			$value .= $ch;
			$ch = &_readChars( $src, $srcType, $encoding, 1 );
		};
		$newPI->nodeValue = $value;
	}
	$newPI;
}

sub _parseElement {
	my ( $src, $srcType, $encoding, $tagName ) = @_;

	if( !&_isLetter( $tagName ) && $tagName ne '_' && $tagName ne ':' )
	{
		$errorMessage = "invalid tag name : ";
		$errorMessage .= &_getCurrentPos( $src, $srcType );
		die $errorMessage;
	}

	my $ch;

	while(1)
	{
		$ch = &_readChars( $src, $srcType, $encoding, 1 );
		last if( &_isSpace($ch) || $ch eq '/' || $ch eq '>' );
		if( !&_isLetter($ch) && !&_isDigit($ch)
			 && $ch ne '.' && $ch ne '-' && $ch ne '_' && $ch ne ':'
			&& !&_isCombiningChar($ch) && !&_isExtender($ch) )
		{
			$errorMessage = "invalid element tagname : ";
			$errorMessage .= &_getCurrentPos( $src, $srcType );
			die $errorMessage;
		}
		$tagName .= $ch;
	}

	my $element = new XML::ESDOM_1_1::Element($tagName);

	if( &_isSpace($ch) ) {
		$ch = &_skipSpace( $src, $srcType, $encoding );
	}
	while( $ch ne '/' && $ch ne '>' )
	{
		&_parseAttribute( $src, $srcType, $encoding, $element, $ch );
		$ch = &_skipSpace( $src, $srcType, $encoding );
	}
	if( $ch eq '/' )
	{
		$ch = &_readChars( $src, $srcType, $encoding, 1 );
		if( $ch eq '>' ) {
			return $element;
		}
		$errorMessage = "invalid element tag : ";
		$errorMessage .= &_getCurrentPos( $src, $srcType );
		die $errorMessage;
	}
	$ch = &_readChars( $src, $srcType, $encoding, 1 );
	while(1)
	{
		if( $ch eq '<' )
		{
			$ch = &_readChars( $src, $srcType, $encoding, 1 );
			if( $ch eq '/' )
			{
				my $tagBuffer;

				$tagBuffer = &_readChars( $src, $srcType, $encoding, length($tagName) );

				$ch = &_skipSpace( $src, $srcType, $encoding );
				if( $tagName ne $tagBuffer || $ch ne '>' )
				{
					$errorMessage = "invalid close tag : ";
					$errorMessage .= &_getCurrentPos( $src, $srcType );
					die $errorMessage;
				}
				return $element;
			}
			elsif( $ch eq '?' )
			{ # Processing instructions
				$element->appendChild( &_parseProcessingInstruction( $src, $srcType, $encoding ) );
			}
			elsif( $ch eq '!' )
			{
				my $name;
				$name = &_readChars( $src, $srcType, $encoding, 2 );
				if( $name eq '--' )
				{ # Comment
					$element->appendChild( &_parseComment( $src, $srcType, $encoding ) );
				}
				elsif( $name eq '[C' )
				{
					$name = &_readChars( $src, $srcType, $encoding, 5 );
					if( $name eq 'DATA[' )
					{ # CDATASection
						$element->appendChild( &_parseCDATASection( $src, $srcType, $encoding ) );
					}
					else
					{
						$errorMessage = "unknown node : ";
						$errorMessage .= &_getCurrentPos( $src, $srcType );
						die $errorMessage;
					}
				}
				else
				{
					$errorMessage = "unknown node : ";
					$errorMessage .= &_getCurrentPos( $src, $srcType );
					die $errorMessage;
				}
			}
			elsif( &_isLetter($ch) || $ch eq '_' || $ch eq ':' )
			{
				$element->appendChild( &_parseElement($src,$srcType,$encoding,$ch) );
			}
			else
			{
				$errorMessage = "unknown node : ";
				$errorMessage .= &_getCurrentPos( $src, $srcType );
				die $errorMessage;
			}
		}
		else
		{
			my $textdata;
			
			do
			{
				if( $ch eq '&' ) {
					$textdata .= &_entityReference2char( $src, $srcType, $encoding );
				}
				else {
					$textdata .= $ch;
				}
				$ch = &_readChars( $src, $srcType, $encoding, 1 );
			}while( $ch ne '<' );
			my $text = new XML::ESDOM_1_1::Text;
			$text->nodeValue = $textdata;
			$element->appendChild( $text );
			next;
		}
		$ch = &_readChars( $src, $srcType, $encoding, 1 );
	}
	return 0;
}

sub _parseAttribute {
	my ( $src, $srcType, $encoding, $element, $name ) = @_;

	if( !&_isLetter($name) && $name ne '_' && $name ne ':' )
	{
		$errorMessage = "invalid attribute name : ";
		$errorMessage .= &_getCurrentPos( $src, $srcType );
		die $errorMessage;
	}

	my $ch;
	my $quot;
	my $value;

	while(1)
	{
		$ch = &_readChars( $src, $srcType, $encoding, 1 );
		last if( &_isSpace($ch) || $ch eq '=' );

		if( !&_isLetter($ch) && !&_isDigit($ch)
			 && $ch ne '.' && $ch ne '-' && $ch ne '_' && $ch ne ':'
			&& !&_isCombiningChar($ch) && !&_isExtender($ch) )
		{
			$errorMessage = "invalid attribute name : ";
			$errorMessage .= &_getCurrentPos( $src, $srcType );
			die $errorMessage;
		}
		$name .= $ch;
	};

	if( $ch ne '=' ) {
		$ch = &_skipSpace( $src, $srcType, $encoding );
	}
	if( $ch ne '=' )
	{
		$errorMessage = "invalid attribute : ";
		$errorMessage .= &_getCurrentPos( $src, $srcType );
		die $errorMessage;
	}
	$quot = $ch = &_skipSpace( $src, $srcType, $encoding );
	if( $quot ne '\'' && $quot ne '"' )
	{
		$errorMessage = "invalid attribute value : ";
		$errorMessage .= &_getCurrentPos( $src, $srcType );
		die $errorMessage;
	}
	while( 1 )
	{
		$ch = &_readChars( $src, $srcType, $encoding, 1 );
		last if( $ch eq $quot );
		if( $ch eq '<' )
		{
			$errorMessage = "invalid attribute value : ";
			$errorMessage .= &_getCurrentPos( $src, $srcType );
			die $errorMessage;
		}
		if( $ch eq '&' ) {
			$value .= &_entityReference2char( $src, $srcType, $encoding );
		}
		else {
			$value .= $ch;
		}
	}

	$element->setAttribute( $name, $value );
}

sub _parseComment {
	my ( $src, $srcType, $encoding ) = @_;

	my $data;
	my $end;

	$end = &_readChars( $src, $srcType, $encoding, 3 );

	while(1)
	{
		last if( $end eq '-->' );
		$data .= substr( $end, 0, 1 );
		$end = substr( $end, 1 ) . &_readChars( $src, $srcType, $encoding, 1 );
	}
	my $comment = new XML::ESDOM_1_1::Comment;
	$comment->nodeValue = $data;
	$comment;
}

sub _parseCDATASection {
	my ( $src, $srcType, $encoding ) = @_;

	my $data;
	my $end;

	$end = &_readChars( $src, $srcType, $encoding, 3 );

	while(1)
	{
		last if( $end eq ']]>' );
		$data .= substr( $end, 0, 1 );
		$end = substr( $end, 1 ) . &_readChars( $src, $srcType, $encoding, 1 );
	}
	my $cdata = new XML::ESDOM_1_1::CDATASection;
	$cdata->nodeValue = $data;
	$cdata;
}

sub _entityReference2char {
	my ( $src, $srcType, $encoding ) = @_;

	my $ch = &_readChars( $src, $srcType, $encoding, 1 );
	if( $ch eq '#' )
	{
		$ch = &_readChars( $src, $srcType, $encoding, 1 );;
		if( $ch eq 'x' )
		{
			my $num = 0;
			my $length = 0;
			while(1)
			{
				last if( $length > 4 );
				$ch = unpack( "U", &_readChars( $src, $srcType, $encoding, 1 ) );
				if( $ch == 0x3b ) {# ';'
					last;
				} elsif( 0x30 <= $ch && $ch <= 0x39 ) { # 0-9
					$num = $num*16 + ( $ch - 0x30 );
				} elsif( 0x41 <= $ch && $ch <= 0x46 ) { # A-F
					$num = $num*16 + ( $ch - 0x41 +10 );
				} elsif( 0x61 <= $ch && $ch <= 0x66 ) { # a-f
					$num = $num*16 + ( $ch - 0x61 +10 );
				} else {
					$errorMessage = "invalid EntityReference : ";
					$errorMessage .= &_getCurrentPos( $src, $srcType );
				}
				$length++;
			}
			return pack( "U", $num );
		}
		else
		{
			my $num = 0;
			my $length = 0;
			while(1)
			{
				last if( $length > 5 );
				$ch = &_readChars( $src, $srcType, $encoding, 1 );
				$ch = unpack( "U", &_readChars( $src, $srcType, $encoding, 1 ) );
				if( $ch == 0x3b ) { # ';'
					last;
				} elsif( 0x30 <= $ch && $ch <= 0x39 ) { # 0-9
					$num = $num*16 + ( $ch - 0x30 );
				} else {
					$errorMessage = "invalid EntityReference : ";
					$errorMessage .= &_getCurrentPos( $src, $srcType );
				}
				$length++;
			}
			return pack( "U", $num );
		}
	}
	elsif( &_isLetter($ch) || $ch eq '_' || $ch eq ':' )
	{
		my $name = $ch;
		while(1)
		{
			$ch = &_readChars( $src, $srcType, $encoding, 1 );
			last if( $ch eq ';' );
			if( !&_isLetter($ch) && !&_isDigit($ch)
			 && $ch ne '.' && $ch ne '-' && $ch ne '_' && $ch ne ':'
			&& !&_isCombiningChar($ch) && !&_isExtender($ch) )
			{
				$errorMessage = "invalid EntityReference : ";
				$errorMessage .= &_getCurrentPos( $src, $srcType );
				die $errorMessage;
			}
			$name .= $ch;
		}
		if( $entityTable{$name} ) {
			return $entityTable{$name};
		}
		unless( $entityTable{'nbsp'} ) {
			$entityTable{'nbsp'} = pack( "U", 160 );
			$entityTable{'iexcl'} = pack( "U", 161 );
			$entityTable{'cent'} = pack( "U", 162 );
			$entityTable{'pound'} = pack( "U", 163 );
			$entityTable{'curren'} = pack( "U", 164 );
			$entityTable{'yen'} = pack( "U", 165 );
			$entityTable{'brvbar'} = pack( "U", 166 );
			$entityTable{'sect'} = pack( "U", 167 );
			$entityTable{'uml'} = pack( "U", 168 );
			$entityTable{'copy'} = pack( "U", 169 );
			$entityTable{'ordf'} = pack( "U", 170 );
			$entityTable{'laquo'} = pack( "U", 171 );
			$entityTable{'not'} = pack( "U", 172 );
			$entityTable{'shy'} = pack( "U", 173 );
			$entityTable{'reg'} = pack( "U", 174 );
			$entityTable{'macr'} = pack( "U", 175 );
			$entityTable{'deg'} = pack( "U", 176 );
			$entityTable{'plusmn'} = pack( "U", 177 );
			$entityTable{'sup2'} = pack( "U", 178 );
			$entityTable{'sup3'} = pack( "U", 179 );
			$entityTable{'acute'} = pack( "U", 180 );
			$entityTable{'micro'} = pack( "U", 181 );
			$entityTable{'para'} = pack( "U", 182 );
			$entityTable{'middot'} = pack( "U", 183 );
			$entityTable{'cedil'} = pack( "U", 184 );
			$entityTable{'sup1'} = pack( "U", 185 );
			$entityTable{'ordm'} = pack( "U", 186 );
			$entityTable{'raquo'} = pack( "U", 187 );
			$entityTable{'frac14'} = pack( "U", 188 );
			$entityTable{'frac12'} = pack( "U", 189 );
			$entityTable{'frac34'} = pack( "U", 190 );
			$entityTable{'iquest'} = pack( "U", 191 );
			$entityTable{'Agrave'} = pack( "U", 192 );
			$entityTable{'Aacute'} = pack( "U", 193 );
			$entityTable{'Acirc'} = pack( "U", 194 );
			$entityTable{'Atilde'} = pack( "U", 195 );
			$entityTable{'Auml'} = pack( "U", 196 );
			$entityTable{'Aring'} = pack( "U", 197 );
			$entityTable{'AElig'} = pack( "U", 198 );
			$entityTable{'Ccedil'} = pack( "U", 199 );
			$entityTable{'Egrave'} = pack( "U", 200 );
			$entityTable{'Eacute'} = pack( "U", 201 );
			$entityTable{'Ecirc'} = pack( "U", 202 );
			$entityTable{'Euml'} = pack( "U", 203 );
			$entityTable{'Igrave'} = pack( "U", 204 );
			$entityTable{'Iacute'} = pack( "U", 205 );
			$entityTable{'Icirc'} = pack( "U", 206 );
			$entityTable{'Iuml'} = pack( "U", 207 );
			$entityTable{'ETH'} = pack( "U", 208 );
			$entityTable{'Ntilde'} = pack( "U", 209 );
			$entityTable{'Ograve'} = pack( "U", 210 );
			$entityTable{'Oacute'} = pack( "U", 211 );
			$entityTable{'Ocirc'} = pack( "U", 212 );
			$entityTable{'Otilde'} = pack( "U", 213 );
			$entityTable{'Ouml'} = pack( "U", 214 );
			$entityTable{'times'} = pack( "U", 215 );
			$entityTable{'Oslash'} = pack( "U", 216 );
			$entityTable{'Ugrave'} = pack( "U", 217 );
			$entityTable{'Uacute'} = pack( "U", 218 );
			$entityTable{'Ucirc'} = pack( "U", 219 );
			$entityTable{'Uuml'} = pack( "U", 220 );
			$entityTable{'Yacute'} = pack( "U", 221 );
			$entityTable{'THORN'} = pack( "U", 222 );
			$entityTable{'szlig'} = pack( "U", 223 );
			$entityTable{'agrave'} = pack( "U", 224 );
			$entityTable{'aacute'} = pack( "U", 225 );
			$entityTable{'acirc'} = pack( "U", 226 );
			$entityTable{'atilde'} = pack( "U", 227 );
			$entityTable{'auml'} = pack( "U", 228 );
			$entityTable{'aring'} = pack( "U", 229 );
			$entityTable{'aelig'} = pack( "U", 230 );
			$entityTable{'ccedil'} = pack( "U", 231 );
			$entityTable{'egrave'} = pack( "U", 232 );
			$entityTable{'eacute'} = pack( "U", 233 );
			$entityTable{'ecirc'} = pack( "U", 234 );
			$entityTable{'euml'} = pack( "U", 235 );
			$entityTable{'igrave'} = pack( "U", 236 );
			$entityTable{'iacute'} = pack( "U", 237 );
			$entityTable{'icirc'} = pack( "U", 238 );
			$entityTable{'iuml'} = pack( "U", 239 );
			$entityTable{'eth'} = pack( "U", 240 );
			$entityTable{'ntilde'} = pack( "U", 241 );
			$entityTable{'ograve'} = pack( "U", 242 );
			$entityTable{'oacute'} = pack( "U", 243 );
			$entityTable{'ocirc'} = pack( "U", 244 );
			$entityTable{'otilde'} = pack( "U", 245 );
			$entityTable{'ouml'} = pack( "U", 246 );
			$entityTable{'divide'} = pack( "U", 247 );
			$entityTable{'oslash'} = pack( "U", 248 );
			$entityTable{'ugrave'} = pack( "U", 249 );
			$entityTable{'uacute'} = pack( "U", 250 );
			$entityTable{'ucirc'} = pack( "U", 251 );
			$entityTable{'uuml'} = pack( "U", 252 );
			$entityTable{'yacute'} = pack( "U", 253 );
			$entityTable{'thorn'} = pack( "U", 254 );
			$entityTable{'yuml'} = pack( "U", 255 );
			$entityTable{'OElig'} = pack( "U", 338 );
			$entityTable{'oelig'} = pack( "U", 339 );
			$entityTable{'Scaron'} = pack( "U", 352 );
			$entityTable{'scaron'} = pack( "U", 353 );
			$entityTable{'Yuml'} = pack( "U", 376 );
			$entityTable{'circ'} = pack( "U", 710 );
			$entityTable{'tilde'} = pack( "U", 732 );
			$entityTable{'ensp'} = pack( "U", 8194 );
			$entityTable{'emsp'} = pack( "U", 8195 );
			$entityTable{'thinsp'} = pack( "U", 8201 );
			$entityTable{'zwnj'} = pack( "U", 8204 );
			$entityTable{'zwj'} = pack( "U", 8205 );
			$entityTable{'lrm'} = pack( "U", 8206 );
			$entityTable{'rlm'} = pack( "U", 8207 );
			$entityTable{'ndash'} = pack( "U", 8211 );
			$entityTable{'mdash'} = pack( "U", 8212 );
			$entityTable{'lsquo'} = pack( "U", 8216 );
			$entityTable{'rsquo'} = pack( "U", 8217 );
			$entityTable{'sbquo'} = pack( "U", 8218 );
			$entityTable{'ldquo'} = pack( "U", 8220 );
			$entityTable{'rdquo'} = pack( "U", 8221 );
			$entityTable{'bdquo'} = pack( "U", 8222 );
			$entityTable{'dagger'} = pack( "U", 8224 );
			$entityTable{'Dagger'} = pack( "U", 8225 );
			$entityTable{'permil'} = pack( "U", 8240 );
			$entityTable{'lsaquo'} = pack( "U", 8249 );
			$entityTable{'rsaquo'} = pack( "U", 8250 );
			$entityTable{'euro'} = pack( "U", 8364 );
			$entityTable{'fnof'} = pack( "U", 402 );
			$entityTable{'Alpha'} = pack( "U", 913 );
			$entityTable{'Beta'} = pack( "U", 914 );
			$entityTable{'Gamma'} = pack( "U", 915 );
			$entityTable{'Delta'} = pack( "U", 916 );
			$entityTable{'Epsilon'} = pack( "U", 917 );
			$entityTable{'Zeta'} = pack( "U", 918 );
			$entityTable{'Eta'} = pack( "U", 919 );
			$entityTable{'Theta'} = pack( "U", 920 );
			$entityTable{'Iota'} = pack( "U", 921 );
			$entityTable{'Kappa'} = pack( "U", 922 );
			$entityTable{'Lambda'} = pack( "U", 923 );
			$entityTable{'Mu'} = pack( "U", 924 );
			$entityTable{'Nu'} = pack( "U", 925 );
			$entityTable{'Xi'} = pack( "U", 926 );
			$entityTable{'Omicron'} = pack( "U", 927 );
			$entityTable{'Pi'} = pack( "U", 928 );
			$entityTable{'Rho'} = pack( "U", 929 );
			$entityTable{'Sigma'} = pack( "U", 931 );
			$entityTable{'Tau'} = pack( "U", 932 );
			$entityTable{'Upsilon'} = pack( "U", 933 );
			$entityTable{'Phi'} = pack( "U", 934 );
			$entityTable{'Chi'} = pack( "U", 935 );
			$entityTable{'Psi'} = pack( "U", 936 );
			$entityTable{'Omega'} = pack( "U", 937 );
			$entityTable{'alpha'} = pack( "U", 945 );
			$entityTable{'beta'} = pack( "U", 946 );
			$entityTable{'gamma'} = pack( "U", 947 );
			$entityTable{'delta'} = pack( "U", 948 );
			$entityTable{'epsilon'} = pack( "U", 949 );
			$entityTable{'zeta'} = pack( "U", 950 );
			$entityTable{'eta'} = pack( "U", 951 );
			$entityTable{'theta'} = pack( "U", 952 );
			$entityTable{'iota'} = pack( "U", 953 );
			$entityTable{'kappa'} = pack( "U", 954 );
			$entityTable{'lambda'} = pack( "U", 955 );
			$entityTable{'mu'} = pack( "U", 956 );
			$entityTable{'nu'} = pack( "U", 957 );
			$entityTable{'xi'} = pack( "U", 958 );
			$entityTable{'omicron'} = pack( "U", 959 );
			$entityTable{'pi'} = pack( "U", 960 );
			$entityTable{'rho'} = pack( "U", 961 );
			$entityTable{'sigmaf'} = pack( "U", 962 );
			$entityTable{'sigma'} = pack( "U", 963 );
			$entityTable{'tau'} = pack( "U", 964 );
			$entityTable{'upsilon'} = pack( "U", 965 );
			$entityTable{'phi'} = pack( "U", 966 );
			$entityTable{'chi'} = pack( "U", 967 );
			$entityTable{'psi'} = pack( "U", 968 );
			$entityTable{'omega'} = pack( "U", 969 );
			$entityTable{'thetasym'} = pack( "U", 977 );
			$entityTable{'upsih'} = pack( "U", 978 );
			$entityTable{'piv'} = pack( "U", 982 );
			$entityTable{'bull'} = pack( "U", 8226 );
			$entityTable{'hellip'} = pack( "U", 8230 );
			$entityTable{'prime'} = pack( "U", 8242 );
			$entityTable{'Prime'} = pack( "U", 8243 );
			$entityTable{'oline'} = pack( "U", 8254 );
			$entityTable{'frasl'} = pack( "U", 8260 );
			$entityTable{'weierp'} = pack( "U", 8472 );
			$entityTable{'image'} = pack( "U", 8465 );
			$entityTable{'real'} = pack( "U", 8476 );
			$entityTable{'trade'} = pack( "U", 8482 );
			$entityTable{'talefsym'} = pack( "U", 8501 );
			$entityTable{'larr'} = pack( "U", 8592 );
			$entityTable{'uarr'} = pack( "U", 8593 );
			$entityTable{'rarr'} = pack( "U", 8594 );
			$entityTable{'darr'} = pack( "U", 8595 );
			$entityTable{'harr'} = pack( "U", 8596 );
			$entityTable{'crarr'} = pack( "U", 8629 );
			$entityTable{'lArr'} = pack( "U", 8656 );
			$entityTable{'uArr'} = pack( "U", 8657 );
			$entityTable{'rArr'} = pack( "U", 8658 );
			$entityTable{'dArr'} = pack( "U", 8659 );
			$entityTable{'hArr'} = pack( "U", 8660 );
			$entityTable{'forall'} = pack( "U", 8704 );
			$entityTable{'part'} = pack( "U", 8706 );
			$entityTable{'exist'} = pack( "U", 8707 );
			$entityTable{'empty'} = pack( "U", 8709 );
			$entityTable{'nabla'} = pack( "U", 8711 );
			$entityTable{'isin'} = pack( "U", 8712 );
			$entityTable{'notin'} = pack( "U", 8713 );
			$entityTable{'ni'} = pack( "U", 8715 );
			$entityTable{'prod'} = pack( "U", 8719 );
			$entityTable{'sum'} = pack( "U", 8721 );
			$entityTable{'minus'} = pack( "U", 8722 );
			$entityTable{'lowast'} = pack( "U", 8727 );
			$entityTable{'radic'} = pack( "U", 8730 );
			$entityTable{'prop'} = pack( "U", 8733 );
			$entityTable{'infin'} = pack( "U", 8734 );
			$entityTable{'ang'} = pack( "U", 8736 );
			$entityTable{'and'} = pack( "U", 8743 );
			$entityTable{'or'} = pack( "U", 8744 );
			$entityTable{'cap'} = pack( "U", 8745 );
			$entityTable{'cup'} = pack( "U", 8746 );
			$entityTable{'int'} = pack( "U", 8747 );
			$entityTable{'there4'} = pack( "U", 8756 );
			$entityTable{'sim'} = pack( "U", 8764 );
			$entityTable{'cong'} = pack( "U", 8773 );
			$entityTable{'asymp'} = pack( "U", 8776 );
			$entityTable{'ne'} = pack( "U", 8800 );
			$entityTable{'equiv'} = pack( "U", 8801 );
			$entityTable{'le'} = pack( "U", 8804 );
			$entityTable{'ge'} = pack( "U", 8805 );
			$entityTable{'sub'} = pack( "U", 8834 );
			$entityTable{'sup'} = pack( "U", 8835 );
			$entityTable{'nsub'} = pack( "U", 8836 );
			$entityTable{'sube'} = pack( "U", 8838 );
			$entityTable{'supe'} = pack( "U", 8839 );
			$entityTable{'oplus'} = pack( "U", 8853 );
			$entityTable{'otimes'} = pack( "U", 8855 );
			$entityTable{'perp'} = pack( "U", 8869 );
			$entityTable{'sdot'} = pack( "U", 8901 );
			$entityTable{'lceil'} = pack( "U", 8968 );
			$entityTable{'rceil'} = pack( "U", 8969 );
			$entityTable{'lfloor'} = pack( "U", 8970 );
			$entityTable{'rfloor'} = pack( "U", 8971 );
			$entityTable{'lang'} = pack( "U", 9001 );
			$entityTable{'rang'} = pack( "U", 9002 );
			$entityTable{'loz'} = pack( "U", 9674 );
			$entityTable{'spades'} = pack( "U", 9824 );
			$entityTable{'clubs'} = pack( "U", 9827 );
			$entityTable{'hearts'} = pack( "U", 9829 );
			$entityTable{'diams'} = pack( "U", 9830 );
		}
		if( $entityTable{$name} ) {
			return $entityTable{$name};
		}
		return '&'.$name.';';
	}
	else
	{
		$errorMessage = "invalid EntityReference : ";
		$errorMessage .= &_getCurrentPos( $src, $srcType );
		die $errorMessage;
	}
}

sub _skipSpace {
	my ( $src, $srcType, $encoding ) = @_;
	my $ch;
	do{
		$ch = &_readChars( $src, $srcType, $encoding, 1 );
	}while( &_isSpace($ch) );
	return $ch;
}

###############################################################################
#

sub _serialize {
	my ( $document, $encoding, $dst, $dstType ) = @_;

	unless( $document->_valid ) {
		$errorMessage = "document is invalid";
		return 0;
	}

	my @bom;
	if( $encoding == UTF_16BE ) {
		$bom[0] = 0xfe;
		$bom[1] = 0xff;
		&_write( $dst, $dstType, \@bom, 2 );
	} elsif( $encoding == UTF_16LE ) {
		$bom[0] = 0xff;
		$bom[1] = 0xfe;
		&_write( $dst, $dstType, \@bom, 2 );
	} elsif( $encoding == UTF_16BE_NONBOM ) {
		$encoding = UTF_16BE;
	} elsif( $encoding == UTF_16LE_NONBOM ) {
		$encoding = UTF_16LE;
	}

	eval{
		&_writeChars( $dst, $dstType, $encoding, '<?xml version="' );
		&_writeChars( $dst, $dstType, $encoding, $document->_version );
		&_writeChars( $dst, $dstType, $encoding, '" encoding="' );
		if( UTF_16BE <= $encoding && $encoding <= UTF_16LE_NONBOM ) {
			&_writeChars( $dst, $dstType, $encoding, 'UTF-16"' );
		} else {
			&_writeChars( $dst, $dstType, $encoding, 'UTF-8"' );
		}
		if( $document->standalone > 0 ) {
			&_writeChars( $dst, $dstType, $encoding, ' standalone="yes"' );
		}
		if( $document->standalone == 0 ) {
			&_writeChars( $dst, $dstType, $encoding, ' standalone="no"' );
		}
		&_writeChars( $dst, $dstType, $encoding, "?>\n" );

		for( my ( $i, $length ) = ( 0, $document->childNodes->length ); $i < $length; $i++ )
		{
			my $node = $document->childNodes->item($i);
			if( $node->nodeType == XML::ESDOM_1_1::Node::PROCESSING_INSTRUCTION_NODE )
			{
				&_writeChars( $dst, $dstType, $encoding, "<?" );
				&_writeChars( $dst, $dstType, $encoding, $node->nodeName );
				&_writeChars( $dst, $dstType, $encoding, " " );
				&_writeChars( $dst, $dstType, $encoding, $node->nodeValue );
				&_writeChars( $dst, $dstType, $encoding, "?>\n" );
			}
			elsif( $node->nodeType == XML::ESDOM_1_1::Node::DOCUMENT_TYPE_NODE )
			{
				&_writeChars( $dst, $dstType, $encoding, "<!DOCTYPE " );
				&_writeChars( $dst, $dstType, $encoding, $node->nodeName );
				&_writeChars( $dst, $dstType, $encoding, ">\n" );
			}
			elsif( $node->nodeType == XML::ESDOM_1_1::Node::COMMENT_NODE )
			{
				&_writeChars( $dst, $dstType, $encoding, "<!-- " );
				&_writeChars( $dst, $dstType, $encoding, $node->nodeValue );
				&_writeChars( $dst, $dstType, $encoding, "-->\n" );
			}
			elsif( $node->nodeType == XML::ESDOM_1_1::Node::ELEMENT_NODE ) {
				&_serializeElement( $document, $encoding, $dst, $dstType, $node );
			}
		}
	};
	if($@) {
		print STDERR $@;
		return 0;
	}
	return 1;
}

sub _serializeElement {
	my ( $document, $encoding, $dst, $dstType, $element ) = @_;

	my $tagName = $element->nodeName;

	&_writeChars( $dst, $dstType, $encoding, "<" );
	&_writeChars( $dst, $dstType, $encoding, $tagName );

	my $attributes = $element->attributes;
	for( my ( $i, $length ) = ( 0, $attributes->length ); $i < $length; $i++ ) {
		&_writeChars( $dst, $dstType, $encoding, ' ' );
		&_writeChars( $dst, $dstType, $encoding, $attributes->item($i)->nodeName );
		&_writeChars( $dst, $dstType, $encoding, '="' );
		&_writeChars( $dst, $dstType, $encoding, _char2entityReference( $attributes->item($i)->nodeValue ) );
		&_writeChars( $dst, $dstType, $encoding, '"' );
	}
	if( !$element->hasChildNodes() ) {
		my $doctype = $document->doctype;
		if( $doctype && $doctype->nodeName =~ /^html/ ) {
			if( $tagName eq "area" || $tagName eq "base" || $tagName eq "br"
				|| $tagName eq "col" || $tagName eq "frame" || $tagName eq "hr"
				|| $tagName eq "img" || $tagName eq "input" || $tagName eq "link"
				|| $tagName eq "meta" ) {
				&_writeChars( $dst, $dstType, $encoding, " />" );
			} else {
				&_writeChars( $dst, $dstType, $encoding, "></" );
				&_writeChars( $dst, $dstType, $encoding, $tagName );
				&_writeChars( $dst, $dstType, $encoding, ">" );
			}
		} else {
			&_writeChars( $dst, $dstType, $encoding, "/>" );
		}
		return;
	}
	&_writeChars( $dst, $dstType, $encoding, ">" );

	my $childNodes = $element->childNodes;
	for( my ( $i, $length ) = ( 0, $childNodes->length ); $i < $length; $i++ )
	{
		my $buffer;

		if( $childNodes->item($i)->nodeType == XML::ESDOM_1_1::Node::ELEMENT_NODE ) {
			&_serializeElement( $document, $encoding, $dst, $dstType, $childNodes->item($i) );
		} elsif ( $childNodes->item($i)->nodeType == XML::ESDOM_1_1::Node::TEXT_NODE ) {
			&_writeChars( $dst, $dstType, $encoding, &_char2entityReference( $childNodes->item($i)->nodeValue ) );
		} elsif ( $childNodes->item($i)->nodeType == XML::ESDOM_1_1::Node::COMMENT_NODE ) {
			&_writeChars( $dst, $dstType, $encoding, "<!--" );
			$buffer = $childNodes->item($i)->nodeValue;
			$buffer =~ s/-->/-- >/go;
			if( $buffer =~ /-$/o ) {
				$buffer .= ' ';
			}
			&_writeChars( $dst, $dstType, $encoding, $buffer );
			&_writeChars( $dst, $dstType, $encoding, "-->" );
		} elsif ( $childNodes->item($i)->nodeType == XML::ESDOM_1_1::Node::CDATA_SECTION_NODE ) {
			&_writeChars( $dst, $dstType, $encoding, "<![CDATA[" );
			$buffer = $childNodes->item($i)->nodeValue;
			$buffer =~ s/\]\]>/]] >/go;
			&_writeChars( $dst, $dstType, $encoding, $buffer );
			&_writeChars( $dst, $dstType, $encoding, "]]>" );
		}
	}
	&_writeChars( $dst, $dstType, $encoding, "</" );
	&_writeChars( $dst, $dstType, $encoding, $tagName );
	&_writeChars( $dst, $dstType, $encoding, ">" );
}

sub _char2entityReference {
	my $str = shift;

	$str =~ s/&/&amp;/go;
	$str =~ s/</&lt;/go;
	$str =~ s/>/&gt;/go;
	$str =~ s/'/&apos;/go;
	$str =~ s/"/&quot;/go;
	$str;
}

##############################################################################
##

sub _readChars {
	my ( $src, $srcType, $encoding, $length ) = @_;

	my @chars;
	my $retStr;

	if( $encoding == UTF_8 ) {
		for( ; $length; $length-- )
		{
			$chars[0] = [&_read( $src, $srcType, 1 )]->[0];
			if( $chars[0] < 0x80 )
			{
				$retStr .= pack( "U", $chars[0] );
				next;
			}

			if( $chars[0] < 0xc2 ) {
				$errorMessage = "invalid character";
				die $errorMessage;
			}

			$chars[1] = [&_read( $src, $srcType, 1 )]->[0];
			if( $chars[0] < 0xe0 )
			{
				if( $chars[1] < 0x80 || 0xbf < $chars[1] ){
					$errorMessage = "invalid character";
					die $errorMessage;
				}
				$retStr .= pack( "U",  ( ( $chars[0] & 0x1f ) << 6 ) + ( $chars[1] & 0x3f ) );
				next;
			}

			$chars[2] = [&_read( $src, $srcType, 1 )]->[0];
			if( $chars[0] < 0xf0 )
			{
				if( ( $chars[0] == 0xe0 && $chars[1] < 0xa0 ) || ( $chars[0] != 0xe0 && $chars[1] < 0x80 ) ) {
					$errorMessage = "invalid character";
					die $errorMessage;
				}
				if( 0xbf < $chars[1] || $chars[2] < 0x80 || 0xbf < $chars[2] ) {
					$errorMessage = "invalid character";
					die $errorMessage;
				}
				$retStr .= pack( "U", ( ( $chars[0] & 0x0f ) << 12 ) + ( ( $chars[1] & 0x3f ) << 6 ) + ( $chars[2] & 0x3f ) );
				next;
			}

			$errorMessage = "invalid character";
			die $errorMessage;
		}
		return $retStr;
	} elsif( $encoding == UTF_16BE ) {
		for( ; $length; $length-- )
		{
			@chars = &_read( $src, $srcType, 2 );
			$retStr .= pack( "U", $chars[0]*0x100 + $chars[1] );
		}
		return $retStr;
	} elsif( $encoding == UTF_16LE ) {
		for( ; $length; $length-- )
		{
			@chars = &_read( $src, $srcType, 2 );
			$retStr .= pack( "U", $chars[1]*0x100 + $chars[0] );
		}
		return $retStr;
	}

	$errorMessage = "invalid encording";
	die $errorMessage;
}

sub _writeChars {
	my ( $dst, $dstType, $encoding, $string ) = @_;

	my @chars;
	my $ch;

	if( $] < 5.006001 ) {
		($string) = $string =~ /((?s).*)/;
	}
	my $length = length($string);

	if( $encoding == UTF_8 ) {
		for( my $index = 0; $index < $length; $index++ )
		{
			$ch = unpack( "U", substr( $string, $index, 1 ) );
			if( $ch < 0x0080 )
			{
				$chars[0] = $ch;
				&_write( $dst, $dstType, \@chars, 1 );
				next;
			} elsif( $ch < 0x00000800 ) {
				$chars[0] = 0xc0 | ( 0x1f & ( $ch >> 6 ) );
				$chars[1] = 0x80 | ( 0x3f & $ch );
				&_write( $dst, $dstType, \@chars, 2 );
				next;
			} else {# if( $ch < 0x00010000 ) {
				$chars[0] = 0xe0 | ( 0x0f & ( $ch >> 12 ) );
				$chars[1] = 0x80 | ( 0x3f & ( $ch >> 6 ) );
				$chars[2] = 0x80 | ( 0x3f & $ch );
				&_write( $dst, $dstType, \@chars, 3 );
				next;
			}
			die "error";
		}
		return;
	} elsif( $encoding == UTF_16BE ) {
		for( my $index = 0; $index < $length; $index++ )
		{
			$ch = unpack( "U", substr( $string, $index, 1 ) );
			$chars[0] = ( $ch - $ch%0x0100 ) / 0x0100;
			$chars[1] = $ch % 0x0100;
			&_write( $dst, $dstType, \@chars, 2 );
		}
		return;
	} elsif( $encoding == UTF_16LE ) {
		for( my $index = 0; $index < $length; $index++ )
		{
			$ch = unpack( "U", substr( $string, $index, 1 ) );
			$chars[0] = $ch % 0x0100;
			$chars[1] = ( $ch - $ch%0x0100 ) / 0x0100;
			&_write( $dst, $dstType, \@chars, 2 );
		}
		return;
	}
	$errorMessage = 'unknown encoding';
	die $errorMessage;
}

sub _read {
	my ( $src, $srcType, $size ) = @_;
	my $buffer;
	my @chars;

	if( $srcType == TYPE_FILE ) {
		if( read( $src, $buffer, $size ) == $size ) {
			@chars = unpack( "C*", $buffer );
			return @chars;
		}
		$errorMessage = "EOF";
		die $errorMessage;
	} elsif( $srcType == TYPE_BUFFER ) {
		if( $bufferInfo{$src}->[$_BI_LENGTH] - $bufferInfo{$src}->[$_BI_CURRENT] >= $size ) {
			@chars = unpack( "C*", substr( $$src, $bufferInfo{$src}->[$_BI_CURRENT], $size ) );
			$bufferInfo{$src}->[$_BI_CURRENT] += $size;
			return @chars;
		}
		$errorMessage = "EOF";
		die $errorMessage;
	}
	$errorMessage = "unknown source type\n";
	die $errorMessage;
}

sub _write {
	my ( $dst, $dstType, $r_buffer, $size ) = @_;

	if( $dstType == TYPE_FILE ) {
		for( my $index = 0; $index < $size; $index++ ) {
			print $dst pack( "C", $r_buffer->[$index] );
		}
		return;
	} elsif( $dstType == TYPE_BUFFER ) {
		for( my $index = 0; $index < $size; $index++ ) {
			$$dst .= pack( "C", $r_buffer->[$index] );
		}
		return;
	}
	$errorMessage = "unknown destination type\n";
	die $errorMessage;
}

##############################################################################
##

sub _getCurrentPos {
	my ( $target, $type ) = @_;

	if( $type == TYPE_FILE ) {
		return tell( $target );
	} elsif( $type == TYPE_BUFFER ) {
		return $bufferInfo{$target}->[$_BI_CURRENT];
	}
	return 0;
}



##############################################################################
##


sub _isSpace {
	my $ch = unpack( "U", shift );
	if( $ch == 0x20 || $ch == 0x09 || $ch == 0x0D || $ch == 0x0A ) {
		return 1;
	}
	return 0;
}

sub _isChar {
	my $ch = unpack( "U", shift );
	if(( $ch == 0x09 )
	|| ( $ch == 0x0A )
	|| ( $ch == 0x0D )
	|| ( 0x0020 <= $ch && $ch <= 0xD7FF )
	|| ( 0xE000 <= $ch && $ch <= 0xFFFD )
	|| ( 0x10000 <= $ch && $ch <= 0x10FFFF )) {
		return 1;
	}
	return 0;
}

sub _isLetter {
	my $ch = shift;
	return ( &_isBaseChar( $ch ) || &_isIdeographic( $ch ) );
}

sub _isBaseChar {
	my $ch = unpack( "U", shift );
	if( $ch < 0x0BA8 ){
		if( $ch < 0x06E5 ){
			if( $ch < 0x03E2 ){
				if( $ch < 0x01F4 ){
					if( $ch < 0x0134 ){
						if(( 0x0041 <= $ch && $ch <= 0x005A )
						|| ( 0x0061 <= $ch && $ch <= 0x007A )
						|| ( 0x00C0 <= $ch && $ch <= 0x00D6 )
						|| ( 0x00D8 <= $ch && $ch <= 0x00F6 )
						|| ( 0x00F8 <= $ch && $ch <= 0x00FF )
						|| ( 0x0100 <= $ch && $ch <= 0x0131 )) {
							return 1;
						}
					}else{
						if(( 0x0134 <= $ch && $ch <= 0x013E )
						|| ( 0x0141 <= $ch && $ch <= 0x0148 )
						|| ( 0x014A <= $ch && $ch <= 0x017E )
						|| ( 0x0180 <= $ch && $ch <= 0x01C3 )
						|| ( 0x01CD <= $ch && $ch <= 0x01F0 )) {
							return 1;
						}
					}
				}else{
					if( $ch < 0x038C ){
						if(( 0x01F4 <= $ch && $ch <= 0x01F5 )
						|| ( 0x01FA <= $ch && $ch <= 0x0217 )
						|| ( 0x0250 <= $ch && $ch <= 0x02A8 )
						|| ( 0x02BB <= $ch && $ch <= 0x02C1 )
						|| ( $ch == 0x0386 )
						|| ( 0x0388 <= $ch && $ch <= 0x038A )) {
							return 1;
						}
					}else{
						if(( $ch == 0x038C )
						|| ( 0x038E <= $ch && $ch <= 0x03A1 )
						|| ( 0x03A3 <= $ch && $ch <= 0x03CE )
						|| ( 0x03D0 <= $ch && $ch <= 0x03D6 )
						|| ( $ch == 0x03DA )
						|| ( $ch == 0x03DC )
						|| ( $ch == 0x03DE )
						|| ( $ch == 0x03E0 )) {
							return 1;
						}
					}
				}
			}else{
				if( $ch < 0x0531 ){
					if( $ch < 0x04C7 ){
						if(( 0x03E2 <= $ch && $ch <= 0x03F3 )
						|| ( 0x0401 <= $ch && $ch <= 0x040C )
						|| ( 0x040E <= $ch && $ch <= 0x044F )
						|| ( 0x0451 <= $ch && $ch <= 0x045C )
						|| ( 0x045E <= $ch && $ch <= 0x0481 )
						|| ( 0x0490 <= $ch && $ch <= 0x04C4 )) {
							return 1;
						}
					}else{
						if(( 0x04C7 <= $ch && $ch <= 0x04C8 )
						|| ( 0x04CB <= $ch && $ch <= 0x04CC )
						|| ( 0x04D0 <= $ch && $ch <= 0x04EB )
						|| ( 0x04EE <= $ch && $ch <= 0x04F5 )
						|| ( 0x04F8 <= $ch && $ch <= 0x04F9 )) {
							return 1;
						}
					}
				}else{
					if( $ch < 0x0641 ){
						if(( 0x0531 <= $ch && $ch <= 0x0556 )
						|| ( $ch == 0x0559 )
						|| ( 0x0561 <= $ch && $ch <= 0x0586 )
						|| ( 0x05D0 <= $ch && $ch <= 0x05EA )
						|| ( 0x05F0 <= $ch && $ch <= 0x05F2 )
						|| ( 0x0621 <= $ch && $ch <= 0x063A )) {
							return 1;
						}
					}else{
						if(( 0x0641 <= $ch && $ch <= 0x064A )
						|| ( 0x0671 <= $ch && $ch <= 0x06B7 )
						|| ( 0x06BA <= $ch && $ch <= 0x06BE )
						|| ( 0x06C0 <= $ch && $ch <= 0x06CE )
						|| ( 0x06D0 <= $ch && $ch <= 0x06D3 )
						|| ( $ch == 0x06D5 )) {
							return 1;
						}
					}
				}
			}
		}else{
			if( $ch < 0x0A8D ){
				if( $ch < 0x09F0 ){
					if( $ch < 0x0993 ){
						if(( 0x06E5 <= $ch && $ch <= 0x06E6 )
						|| ( 0x0905 <= $ch && $ch <= 0x0939 )
						|| ( $ch == 0x093D )
						|| ( 0x0958 <= $ch && $ch <= 0x0961 )
						|| ( 0x0985 <= $ch && $ch <= 0x098C )
						|| ( 0x098F <= $ch && $ch <= 0x0990 )) {
							return 1;
						}
					}else{
						if(( 0x0993 <= $ch && $ch <= 0x09A8 )
						|| ( 0x09AA <= $ch && $ch <= 0x09B0 )
						|| ( $ch == 0x09B2 )
						|| ( 0x09B6 <= $ch && $ch <= 0x09B9 )
						|| ( 0x09DC <= $ch && $ch <= 0x09DD )
						|| ( 0x09DF <= $ch && $ch <= 0x09E1 )) {
							return 1;
						}
					}
				}else{
					if( $ch < 0x0A35 ){
						if(( 0x09F0 <= $ch && $ch <= 0x09F1 )
						|| ( 0x0A05 <= $ch && $ch <= 0x0A0A )
						|| ( 0x0A0F <= $ch && $ch <= 0x0A10 )
						|| ( 0x0A13 <= $ch && $ch <= 0x0A28 )
						|| ( 0x0A2A <= $ch && $ch <= 0x0A30 )
						|| ( 0x0A32 <= $ch && $ch <= 0x0A33 )) {
							return 1;
						}
					}else{
						if(( 0x0A35 <= $ch && $ch <= 0x0A36 )
						|| ( 0x0A38 <= $ch && $ch <= 0x0A39 )
						|| ( 0x0A59 <= $ch && $ch <= 0x0A5C )
						|| ( $ch == 0x0A5E )
						|| ( 0x0A72 <= $ch && $ch <= 0x0A74 )
						|| ( 0x0A85 <= $ch && $ch <= 0x0A8B )) {
							return 1;
						}
					}
				}
			}else{
				if( $ch < 0x0B32 ){
					if( $ch < 0x0ABD ){
						if(( $ch == 0x0A8D )
						|| ( 0x0A8F <= $ch && $ch <= 0x0A91 )
						|| ( 0x0A93 <= $ch && $ch <= 0x0AA8 )
						|| ( 0x0AAA <= $ch && $ch <= 0x0AB0 )
						|| ( 0x0AB2 <= $ch && $ch <= 0x0AB3 )
						|| ( 0x0AB5 <= $ch && $ch <= 0x0AB9 )) {
							return 1;
						}
					}else{
						if(( $ch == 0x0ABD )
						|| ( $ch == 0x0AE0 )
						|| ( 0x0B05 <= $ch && $ch <= 0x0B0C )
						|| ( 0x0B0F <= $ch && $ch <= 0x0B10 )
						|| ( 0x0B13 <= $ch && $ch <= 0x0B28 )
						|| ( 0x0B2A <= $ch && $ch <= 0x0B30 )) {
							return 1;
						}
					}
				}else{
					if( $ch < 0x0B8E ){
						if(( 0x0B32 <= $ch && $ch <= 0x0B33 )
						|| ( 0x0B36 <= $ch && $ch <= 0x0B39 )
						|| ( $ch == 0x0B3D )
						|| ( 0x0B5C <= $ch && $ch <= 0x0B5D )
						|| ( 0x0B5F <= $ch && $ch <= 0x0B61 )
						|| ( 0x0B85 <= $ch && $ch <= 0x0B8A )) {
							return 1;
						}
					}else{
						if(( 0x0B8E <= $ch && $ch <= 0x0B90 )
						|| ( 0x0B92 <= $ch && $ch <= 0x0B95 )
						|| ( 0x0B99 <= $ch && $ch <= 0x0B9A )
						|| ( $ch == 0x0B9C )
						|| ( 0x0B9E <= $ch && $ch <= 0x0B9F )) {
							return 1;
						}
					}
				}
			}
		}
	}else{
		if( $ch < 0x110E ){
			if( $ch < 0x0E32 ){
				if( $ch < 0x0C92 ){
					if( $ch < 0x0C12 ){
						if(( 0x0BA3 <= $ch && $ch <= 0x0BA4 )
						|| ( 0x0BA8 <= $ch && $ch <= 0x0BAA )
						|| ( 0x0BAE <= $ch && $ch <= 0x0BB5 )
						|| ( 0x0BB7 <= $ch && $ch <= 0x0BB9 )
						|| ( 0x0C05 <= $ch && $ch <= 0x0C0C )
						|| ( 0x0C0E <= $ch && $ch <= 0x0C10 )) {
							return 1;
						}
					}else{
						if(( 0x0C12 <= $ch && $ch <= 0x0C28 )
						|| ( 0x0C2A <= $ch && $ch <= 0x0C33 )
						|| ( 0x0C35 <= $ch && $ch <= 0x0C39 )
						|| ( 0x0C60 <= $ch && $ch <= 0x0C61 )
						|| ( 0x0C85 <= $ch && $ch <= 0x0C8C )
						|| ( 0x0C8E <= $ch && $ch <= 0x0C90 )) {
							return 1;
						}
					}
				}else{
					if( $ch < 0x0D0E ){
						if(( 0x0C92 <= $ch && $ch <= 0x0CA8 )
						|| ( 0x0CAA <= $ch && $ch <= 0x0CB3 )
						|| ( 0x0CB5 <= $ch && $ch <= 0x0CB9 )
						|| ( $ch == 0x0CDE )
						|| ( 0x0CE0 <= $ch && $ch <= 0x0CE1 )
						|| ( 0x0D05 <= $ch && $ch <= 0x0D0C )) {
							return 1;
						}
					}else{
						if(( 0x0D0E <= $ch && $ch <= 0x0D10 )
						|| ( 0x0D12 <= $ch && $ch <= 0x0D28 )
						|| ( 0x0D2A <= $ch && $ch <= 0x0D39 )
						|| ( 0x0D60 <= $ch && $ch <= 0x0D61 )
						|| ( 0x0E01 <= $ch && $ch <= 0x0E2E )
						|| ( $ch == 0x0E30 )) {
							return 1;
						}
					}
				}
			}else{
				if( $ch < 0x0EB0 ){
					if( $ch < 0x0E94 ){
						if(( 0x0E32 <= $ch && $ch <= 0x0E33 )
						|| ( 0x0E40 <= $ch && $ch <= 0x0E45 )
						|| ( 0x0E81 <= $ch && $ch <= 0x0E82 )
						|| ( $ch == 0x0E84 )
						|| ( 0x0E87 <= $ch && $ch <= 0x0E88 )
						|| ( $ch == 0x0E8A )
						|| ( $ch == 0x0E8D )) {
							return 1;
						}
					}else{
						if(( 0x0E94 <= $ch && $ch <= 0x0E97 )
						|| ( 0x0E99 <= $ch && $ch <= 0x0E9F )
						|| ( 0x0EA1 <= $ch && $ch <= 0x0EA3 )
						|| ( $ch == 0x0EA5 )
						|| ( $ch == 0x0EA7 )
						|| ( 0x0EAA <= $ch && $ch <= 0x0EAB )
						|| ( 0x0EAD <= $ch && $ch <= 0x0EAE )) {
							return 1;
						}
					}
				}else{
					if( $ch < 0x10D0 ){
						if(( $ch == 0x0EB0 )
						|| ( 0x0EB2 <= $ch && $ch <= 0x0EB3 )
						|| ( $ch == 0x0EBD )
						|| ( 0x0EC0 <= $ch && $ch <= 0x0EC4 )
						|| ( 0x0F40 <= $ch && $ch <= 0x0F47 )
						|| ( 0x0F49 <= $ch && $ch <= 0x0F69 )
						|| ( 0x10A0 <= $ch && $ch <= 0x10C5 )) {
							return 1;
						}
					}else{
						if(( 0x10D0 <= $ch && $ch <= 0x10F6 )
						|| ( $ch == 0x1100 )
						|| ( 0x1102 <= $ch && $ch <= 0x1103 )
						|| ( 0x1105 <= $ch && $ch <= 0x1107 )
						|| ( $ch == 0x1109 )
						|| ( 0x110B <= $ch && $ch <= 0x110C )) {
							return 1;
						}
					}
				}
			}
		}else{
			if( $ch < 0x1F20 ){
				if( $ch < 0x1175 ){
					if( $ch < 0x115F ){
						if(( 0x110E <= $ch && $ch <= 0x1112 )
						|| ( $ch == 0x113C )
						|| ( $ch == 0x113E )
						|| ( $ch == 0x1140 )
						|| ( $ch == 0x114C )
						|| ( $ch == 0x114E )
						|| ( $ch == 0x1150 )
						|| ( 0x1154 <= $ch && $ch <= 0x1155 )
						|| ( $ch == 0x1159 )) {
							return 1;
						}
					}else{
						if(( 0x115F <= $ch && $ch <= 0x1161 )
						|| ( $ch == 0x1163 )
						|| ( $ch == 0x1165 )
						|| ( $ch == 0x1167 )
						|| ( $ch == 0x1169 )
						|| ( 0x116D <= $ch && $ch <= 0x116E )
						|| ( 0x1172 <= $ch && $ch <= 0x1173 )) {
							return 1;
						}
					}
				}else{
					if( $ch < 0x11BC ){
						if(( $ch == 0x1175 )
						|| ( $ch == 0x119E )
						|| ( $ch == 0x11A8 )
						|| ( $ch == 0x11AB )
						|| ( 0x11AE <= $ch && $ch <= 0x11AF )
						|| ( 0x11B7 <= $ch && $ch <= 0x11B8 )
						|| ( $ch == 0x11BA )) {
							return 1;
						}
					}else{
						if(( 0x11BC <= $ch && $ch <= 0x11C2 )
						|| ( $ch == 0x11EB )
						|| ( $ch == 0x11F0 )
						|| ( $ch == 0x11F9 )
						|| ( 0x1E00 <= $ch && $ch <= 0x1E9B )
						|| ( 0x1EA0 <= $ch && $ch <= 0x1EF9 )
						|| ( 0x1F00 <= $ch && $ch <= 0x1F15 )
						|| ( 0x1F18 <= $ch && $ch <= 0x1F1D )) {
							return 1;
						}
					}
				}
			}else{
				if( $ch < 0x1FD6 ){
					if( $ch < 0x1F80 ){
						if(( 0x1F20 <= $ch && $ch <= 0x1F45 )
						|| ( 0x1F48 <= $ch && $ch <= 0x1F4D )
						|| ( 0x1F50 <= $ch && $ch <= 0x1F57 )
						|| ( $ch == 0x1F59 )
						|| ( $ch == 0x1F5B )
						|| ( $ch == 0x1F5D )
						|| ( 0x1F5F <= $ch && $ch <= 0x1F7D )) {
							return 1;
						}
					}else{
						if(( 0x1F80 <= $ch && $ch <= 0x1FB4 )
						|| ( 0x1FB6 <= $ch && $ch <= 0x1FBC )
						|| ( $ch == 0x1FBE )
						|| ( 0x1FC2 <= $ch && $ch <= 0x1FC4 )
						|| ( 0x1FC6 <= $ch && $ch <= 0x1FCC )
						|| ( 0x1FD0 <= $ch && $ch <= 0x1FD3 )) {
							return 1;
						}
					}
				}else{
					if( $ch < 0x212E ){
						if(( 0x1FD6 <= $ch && $ch <= 0x1FDB )
						|| ( 0x1FE0 <= $ch && $ch <= 0x1FEC )
						|| ( 0x1FF2 <= $ch && $ch <= 0x1FF4 )
						|| ( 0x1FF6 <= $ch && $ch <= 0x1FFC )
						|| ( $ch == 0x2126 )
						|| ( 0x212A <= $ch && $ch <= 0x212B )) {
							return 1;
						}
					}else{
						if(( $ch == 0x212E )
						|| ( 0x2180 <= $ch && $ch <= 0x2182 )
						|| ( 0x3041 <= $ch && $ch <= 0x3094 )
						|| ( 0x30A1 <= $ch && $ch <= 0x30FA )
						|| ( 0x3105 <= $ch && $ch <= 0x312C )
						|| ( 0xAC00 <= $ch && $ch <= 0xD7A3 )) {
							return 1;
						}
					}
				}
			}
		}
	}
	return 0;
}

sub _isIdeographic {
	my $ch = unpack( "U", shift );
	if(( 0x4E00 <= $ch && $ch <= 0x9FA5 )
	|| ( $ch == 0x3007 )
	|| ( 0x3021 <= $ch && $ch <= 0x3029 )) {
		return 1;
	}
	return 0;
}

sub _isCombiningChar {
	my $ch = unpack( "U", shift );
	if( $ch < 0x0B4B ){
		if( $ch < 0x09BC ){
			if( $ch < 0x06DD ){
				if( $ch < 0x05BB){
					if(( 0x0300 <= $ch && $ch <= 0x0345 )
					|| ( 0x0360 <= $ch && $ch <= 0x0361 )
					|| ( 0x0483 <= $ch && $ch <= 0x0486 )
					|| ( 0x0591 <= $ch && $ch <= 0x05A1 )
					|| ( 0x05A3 <= $ch && $ch <= 0x05B9 )) {
						return 1;
					}
				}else{
					if(( 0x05BB <= $ch && $ch <= 0x05BD )
					|| ( $ch == 0x05BF )
					|| ( 0x05C1 <= $ch && $ch <= 0x05C2 )
					|| ( $ch == 0x05C4 )
					|| ( 0x064B <= $ch && $ch <= 0x0652 )
					|| ( $ch == 0x0670 )
					|| ( 0x06D6 <= $ch && $ch <= 0x06DC )) {
						return 1;
					}
				}
			}else{
				if( $ch < 0x093C ){
					if(( 0x06DD <= $ch && $ch <= 0x06DF )
					|| ( 0x06E0 <= $ch && $ch <= 0x06E4 )
					|| ( 0x06E7 <= $ch && $ch <= 0x06E8 )
					|| ( 0x06EA <= $ch && $ch <= 0x06ED )
					|| ( 0x0901 <= $ch && $ch <= 0x0903 )) {
						return 1;
					}
				}else{
					if(( $ch == 0x093C )
					|| ( 0x093E <= $ch && $ch <= 0x094C )
					|| ( $ch == 0x094D )
					|| ( 0x0951 <= $ch && $ch <= 0x0954 )
					|| ( 0x0962 <= $ch && $ch <= 0x0963 )
					|| ( 0x0981 <= $ch && $ch <= 0x0983 )) {
						return 1;
					}
				}
			}
		}else{
			if( $ch < 0x0A4B ){
				if( $ch < 0x09E2 ){
					if(( $ch == 0x09BC )
					|| ( $ch == 0x09BE )
					|| ( $ch == 0x09BF )
					|| ( 0x09C0 <= $ch && $ch <= 0x09C4 )
					|| ( 0x09C7 <= $ch && $ch <= 0x09C8 )
					|| ( 0x09CB <= $ch && $ch <= 0x09CD )
					|| ( $ch == 0x09D7 )) {
						return 1;
					}
				}else{
					if(( 0x09E2 <= $ch && $ch <= 0x09E3 )
					|| ( $ch == 0x0A02 )
					|| ( $ch == 0x0A3C )
					|| ( $ch == 0x0A3E )
					|| ( $ch == 0x0A3F )
					|| ( 0x0A40 <= $ch && $ch <= 0x0A42 )
					|| ( 0x0A47 <= $ch && $ch <= 0x0A48 )) {
						return 1;
					}
				}
			}else{
				if( $ch < 0x0ACB ){
					if(( 0x0A4B <= $ch && $ch <= 0x0A4D )
					|| ( 0x0A70 <= $ch && $ch <= 0x0A71 )
					|| ( 0x0A81 <= $ch && $ch <= 0x0A83 )
					|| ( $ch == 0x0ABC )
					|| ( 0x0ABE <= $ch && $ch <= 0x0AC5 )
					|| ( 0x0AC7 <= $ch && $ch <= 0x0AC9 )) {
						return 1;
					}
				}else{
					if(( 0x0ACB <= $ch && $ch <= 0x0ACD )
					|| ( 0x0B01 <= $ch && $ch <= 0x0B03 )
					|| ( $ch == 0x0B3C )
					|| ( 0x0B3E <= $ch && $ch <= 0x0B43 )
					|| ( 0x0B47 <= $ch && $ch <= 0x0B48 )) {
						return 1;
					}
				}
			}
		}
	}else{
		if( $ch < 0x0D57 ){
			if( $ch < 0x0C55 ){
				if( $ch < 0x0BCA ){
					if(( 0x0B4B <= $ch && $ch <= 0x0B4D )
					|| ( 0x0B56 <= $ch && $ch <= 0x0B57 )
					|| ( 0x0B82 <= $ch && $ch <= 0x0B83 )
					|| ( 0x0BBE <= $ch && $ch <= 0x0BC2 )
					|| ( 0x0BC6 <= $ch && $ch <= 0x0BC8 )) {
						return 1;
					}
				}else{
					if(( 0x0BCA <= $ch && $ch <= 0x0BCD )
					|| ( $ch == 0x0BD7 )
					|| ( 0x0C01 <= $ch && $ch <= 0x0C03 )
					|| ( 0x0C3E <= $ch && $ch <= 0x0C44 )
					|| ( 0x0C46 <= $ch && $ch <= 0x0C48 )
					|| ( 0x0C4A <= $ch && $ch <= 0x0C4D )) {
						return 1;
					}
				}
			}else{
				if( $ch < 0x0CD5 ){
					if(( 0x0C55 <= $ch && $ch <= 0x0C56 )
					|| ( 0x0C82 <= $ch && $ch <= 0x0C83 )
					|| ( 0x0CBE <= $ch && $ch <= 0x0CC4 )
					|| ( 0x0CC6 <= $ch && $ch <= 0x0CC8 )
					|| ( 0x0CCA <= $ch && $ch <= 0x0CCD )) {
						return 1;
					}
				}else{
					if(( 0x0CD5 <= $ch && $ch <= 0x0CD6 )
					|| ( 0x0D02 <= $ch && $ch <= 0x0D03 )
					|| ( 0x0D3E <= $ch && $ch <= 0x0D43 )
					|| ( 0x0D46 <= $ch && $ch <= 0x0D48 )
					|| ( 0x0D4A <= $ch && $ch <= 0x0D4D )) {
						return 1;
					}
				}
			}
		}else{
			if( $ch < 0x0F3F ){
				if( $ch < 0x0EBB ){
					if(( $ch == 0x0D57 )
					|| ( $ch == 0x0E31 )
					|| ( 0x0E34 <= $ch && $ch <= 0x0E3A )
					|| ( 0x0E47 <= $ch && $ch <= 0x0E4E )
					|| ( $ch == 0x0EB1 )
					|| ( 0x0EB4 <= $ch && $ch <= 0x0EB9 )) {
						return 1;
					}
				}else{
					if(( 0x0EBB <= $ch && $ch <= 0x0EBC )
					|| ( 0x0EC8 <= $ch && $ch <= 0x0ECD )
					|| ( 0x0F18 <= $ch && $ch <= 0x0F19 )
					|| ( $ch == 0x0F35 )
					|| ( $ch == 0x0F37 )
					|| ( $ch == 0x0F39 )
					|| ( $ch == 0x0F3E )) {
						return 1;
					}
				}
			}else{
				if( $ch < 0x0FB1 ){
					if(( $ch == 0x0F3F )
					|| ( 0x0F71 <= $ch && $ch <= 0x0F84 )
					|| ( 0x0F86 <= $ch && $ch <= 0x0F8B )
					|| ( 0x0F90 <= $ch && $ch <= 0x0F95 )
					|| ( $ch == 0x0F97 )
					|| ( 0x0F99 <= $ch && $ch <= 0x0FAD )) {
						return 1;
					}
				}else{
					if(( 0x0FB1 <= $ch && $ch <= 0x0FB7 )
					|| ( $ch == 0x0FB9 )
					|| ( 0x20D0 <= $ch && $ch <= 0x20DC )
					|| ( $ch == 0x20E1 )
					|| ( 0x302A <= $ch && $ch <= 0x302F )
					|| ( $ch == 0x3099 )
					|| ( $ch == 0x309A )) {
						return 1;
					}
				}
			}
		}
	}
	return 0;
}

sub _isDigit {
	my $ch = unpack( "U", shift );
	if( $ch < 0x0BE7 ){
		if( $ch < 0x09E6 ){
			if(( 0x0030 <= $ch && $ch <= 0x0039 )
			|| ( 0x0660 <= $ch && $ch <= 0x0669 )
			|| ( 0x06F0 <= $ch && $ch <= 0x06F9 )
			|| ( 0x0966 <= $ch && $ch <= 0x096F )) {
				return 1;
			}
		}else{
			if(( 0x09E6 <= $ch && $ch <= 0x09EF )
			|| ( 0x0A66 <= $ch && $ch <= 0x0A6F )
			|| ( 0x0AE6 <= $ch && $ch <= 0x0AEF )
			|| ( 0x0B66 <= $ch && $ch <= 0x0B6F )) {
				return 1;
			}
		}
	}else{
		if( $ch < 0x0E50 ){
			if(( 0x0BE7 <= $ch && $ch <= 0x0BEF )
			|| ( 0x0C66 <= $ch && $ch <= 0x0C6F )
			|| ( 0x0CE6 <= $ch && $ch <= 0x0CEF )
			|| ( 0x0D66 <= $ch && $ch <= 0x0D6F )) {
				return 1;
			}
		}else{
			if(( 0x0E50 <= $ch && $ch <= 0x0E59 )
			|| ( 0x0ED0 <= $ch && $ch <= 0x0ED9 )
			|| ( 0x0F20 <= $ch && $ch <= 0x0F29 )) {
				return 1;
			}
		}
	}
	return 0;
}

sub _isExtender {
	my $ch = unpack( "U", shift );
	if( $ch < 0x3005 ){
		if(( $ch == 0x00B7 )
		|| ( $ch == 0x02D0 )
		|| ( $ch == 0x02D1 )
		|| ( $ch == 0x0387 )
		|| ( $ch == 0x0640 )
		|| ( $ch == 0x0E46 )
		|| ( $ch == 0x0EC6 )) {
			return 1;
		}
	}else{
		if(( $ch == 0x3005 )
		|| ( 0x3031 <= $ch && $ch <= 0x3035 )
		|| ( 0x309D <= $ch && $ch <= 0x309E )
		|| ( 0x30FC <= $ch && $ch <= 0x30FE )) {
			return 1;
		}
	}
	return 0;
}

1;