##############################################################################
package XML::ESDOM_1_1::Document;
##############################################################################
use utf8;
use strict;
use XML::ESDOM_1_1::Node;
use XML::ESDOM_1_1::Element;
use XML::ESDOM_1_1::Text;
use XML::ESDOM_1_1::Comment;
use XML::ESDOM_1_1::CDATASection;
use XML::ESDOM_1_1::DocumentType;

our @ISA = qw ( XML::ESDOM_1_1::Node );
our ( $__nodeName,
	$__nodeValue,
	$__nodeType,
	$__parentNode,
	$__childNodes,
	$__attributes,
	$__canHaveChild
	) = (
	$XML::ESDOM_1_1::Node::__nodeName,
	$XML::ESDOM_1_1::Node::__nodeValue,
	$XML::ESDOM_1_1::Node::__nodeType,
	$XML::ESDOM_1_1::Node::__parentNode,
	$XML::ESDOM_1_1::Node::__childNodes,
	$XML::ESDOM_1_1::Node::__attributes,
	$XML::ESDOM_1_1::Node::__canHaveChild
	);

our( $__version, $__encoding, $__standalone ) = (0,1,2);

our( $__XMLDecl,
 $__docType,
 $__documentElement,
 $__valid
 ) = (10,11,12,13,14);

sub new {
	my $class = shift;
	my $self = new XML::ESDOM_1_1::Node( '#document' );
	$self->[$__canHaveChild] = 1;

	$self->[$__nodeType] = XML::ESDOM_1_1::Node::DOCUMENT_NODE;

	bless $self;

	$self->[$__XMLDecl] = ();
	$self->[$__XMLDecl][$__version] = '1.0';
	$self->[$__XMLDecl][$__standalone] = -1;
	$self->[$__docType] = '';
	$self->[$__documentElement] = '';
	$self->[$__valid] = 0;
	$self;
}

sub doctype {
	$_[0]->[$__docType];
}

sub documentElement {
	$_[0]->[$__documentElement];
}

sub standalone : lvalue {
	$_[0]->[$__XMLDecl][$__standalone];
}

sub createElement {
	my $self = shift;
	my $tagName = shift;
	new XML::ESDOM_1_1::Element( $tagName );
}
sub createTextNode {
	my $self = shift;
	my $text = shift;
	my $textNode = new XML::ESDOM_1_1::Text;
	$textNode->nodeValue = $text;
	$textNode;
}

sub createComment {
	my $self = shift;
	my $comment = shift;
	my $commentNode = new XML::ESDOM_1_1::Comment;
	$commentNode->nodeValue = $comment;
	$commentNode;
}

sub createCDATASection {
	my $self = shift;
	my $cdata = shift;
	my $cdatasectionNode = new XML::ESDOM_1_1::CDATASection;
	$cdatasectionNode->nodeValue = $cdata;
	$cdatasectionNode;
}

sub getElementsByTagName {
	my $self = shift;
	my $tagName = shift;

	if( $self->documentElement->nodeName ne $tagName && $tagName ne '*' ) {
		return $self->documentElement->getElementsByTagName( $tagName );
	}

	my $nodeList = new XML::ESDOM_1_1::NodeList;
	$nodeList->_addNode( $self->documentElement );

	my $nodeList2 = $self->documentElement->getElementsByTagName( $tagName );
	my $i;
	for( $i = 0; $i < $nodeList2->length; $i++ ) {
		$nodeList->_addNode( $nodeList2->item($i) );
	}
	$nodeList;
}

sub getElementById {
	my $self = shift;
	my $id = shift;
	my $element = shift;
	my $ret;

	unless( $element ) {
		$element = $self->documentElement;
	}

	if( $element->getAttribute('id') eq $id ) {
		return $element;
	}

	for( my $i = 0; $i < $element->childNodes->length; $i++ ) {
		if( $element->childNodes->item($i)->nodeType == XML::ESDOM_1_1::Node::ELEMENT_NODE ) {
			$ret = &getElementById( $self, $id, $element->childNodes->item($i) );
			last if( $ret );
		}
	}
	$ret;
}

sub setDoctype {
	my $self = shift;
	my $doctype = shift;
	my $newDoctype =  new XML::ESDOM_1_1::DocumentType( $doctype );
	$self->[$__docType] = $newDoctype;
	for( my $i = 0; $i < $self->childNodes->length; $i++ ) {
		if( $self->childNodes->item($i)->nodeType == DOCUMENT_TYPE_NODE ) {
			my $old = $self->childNodes->_replaceNode( $newDoctype, $self->childNodes->item($i) );
			if( $old ) {
				return $old->nodeName;
			}
			return;
		}
	}
	$self->childNodes->_addNode( $newDoctype );
	return;
}


#############################
sub _version : lvalue {
	$_[0]->[$__XMLDecl][$__version];
}

sub _setDocumentElement {
	my $self = shift;
	my $element = shift;
	if( $element->parentNode ) {
		return;
	}
	if( $self->[$__documentElement] ) {
		$self->replaceChild( $element, $self->documentElement );
	} else {
		$self->appendChild( $element );
	}
	$self->[$__documentElement] = $element;
}

sub _valid : lvalue {
	$_[0]->[$__valid];
}

1;

