##############################################################################
package XML::ESDOM_1_1::Element;
##############################################################################
use utf8;
use strict;
use XML::ESDOM_1_1::Node;
our @ISA = qw ( XML::ESDOM_1_1::Node );

our ( $__nodeName,
	$__nodeValue,
	$__nodeType,
	$__parentNode,
	$__childNodes,
	$__attributes,
	$__canHaveChild
	) = (
	$XML::ESDOM_1_1::Node::__nodeName,
	$XML::ESDOM_1_1::Node::__nodeValue,
	$XML::ESDOM_1_1::Node::__nodeType,
	$XML::ESDOM_1_1::Node::__parentNode,
	$XML::ESDOM_1_1::Node::__childNodes,
	$XML::ESDOM_1_1::Node::__attributes,
	$XML::ESDOM_1_1::Node::__canHaveChild
	);


sub new {
	my $class = shift;
	my $tagName = shift;
	unless( $tagName ) {
		return;
	}
	my $self = new XML::ESDOM_1_1::Node( $tagName );
	$self->[$__canHaveChild] = 1;
	$self->[$__nodeType] = XML::ESDOM_1_1::Node::ELEMENT_NODE;
	$self->[$__attributes] = new XML::ESDOM_1_1::NamedNodeMap;

	bless $self;
}

sub getAttribute {
	my $self = shift;
	my $name = shift;
	unless( $self->attributes ) { return }
	if( $self->attributes->getNamedItem( $name ) ) {
		return $self->attributes->getNamedItem( $name )->nodeValue;
	}
	return '';
}

sub setAttribute {
	my $self = shift;
	my ( $name, $value ) = @_;
	unless( $self->attributes ) { return }
	my $attr = new XML::ESDOM_1_1::Node( $name, XML::ESDOM_1_1::Node::ATTRIBUTE_NODE );
	$attr->nodeValue = $value;
	$self->attributes->setNamedItem( $attr );
	return;
}

sub removeAttribute {
	my $self = shift;
	my $name = shift;
	unless( $self->attributes ) { return }
	$self->attributes->removeNamedItem( $name );
	return;
}

sub getAttributeNode {
	my $self = shift;
	my $name = shift;
	unless( $self->attributes ) { return }
	$self->attributes->getNamedItem( $name );
}

sub setAttributeNode {
	my $self = shift;
	my $newAttr = shift;
	return if( !$newAttr || $newAttr->nodeType != XML::ESDOM_1_1::Node::ATTRIBUTE_NODE );
	$self->attributes->setNamedItem( $newAttr );
}

sub removeAttributeNode {
	my $self = shift;
	my $oldAttr = shift;
	return if( !$oldAttr || $oldAttr->nodeType != XML::ESDOM_1_1::Node::ATTRIBUTE_NODE );
	$self->attributes->removeNamedItem( $oldAttr->nodeName );
}

sub getElementsByTagName {
	my $self = shift;
	my $tagName = shift;

	my $length = $self->childNodes->length;
	my $i;
	my $nodeList = new XML::ESDOM_1_1::NodeList;
	for( $i = 0; $i < $length; $i++ ) {
		my $childNode = $self->childNodes->item( $i );
		if( $childNode->nodeType == XML::ESDOM_1_1::Node::ELEMENT_NODE ) {
			if( $childNode->nodeName eq $tagName || $tagName eq '*' ) {
				$nodeList->_addNode( $childNode );
			}
			my $nodeList2 = $childNode->getElementsByTagName( $tagName );
			my $j;
			for( $j = 0; $j < $nodeList2->length; $j++ ) {
				$nodeList->_addNode( $nodeList2->item($j) );
			}
		}
	}
	$nodeList;
}

sub normalize {
	my $self = shift;
	for( my $i = 0; $i < $self->childNodes->length; ) {
		if( $self->childNodes->item($i)->nodeType == XML::ESDOM_1_1::Node::ELEMENT_NODE ) {
			$self->childNodes->item($i)->normalize();
		} elsif( $self->childNodes->item($i)->nodeType == XML::ESDOM_1_1::Node::TEXT_NODE ) {
			if( $self->childNodes->item($i+1)
					&& $self->childNodes->item($i+1)->nodeType == XML::ESDOM_1_1::Node::TEXT_NODE ) {
				$self->childNodes->item($i)->nodeValue
				 = $self->childNodes->item($i)->nodeValue.$self->childNodes->item($i+1)->nodeValue;
				$self->removeChild( $self->childNodes->item($i+1) );
				next;
			}
		}
		$i++;
	}
}

1;

