##############################################################################
package XML::ESDOM_1_1::Node;
##############################################################################
use strict;

use XML::ESDOM_1_1::NodeList;
use XML::ESDOM_1_1::NamedNodeMap;
require Exporter;
our(@ISA, @EXPORT);

@ISA = qw(Exporter);
@EXPORT = qw(
	ELEMENT_NODE
	ATTRIBUTE_NODE
	TEXT_NODE
	CDATA_SECTION_NODE
	ENTITY_REFERENCE_NODE
	ENTITY_NODE
	PROCESSING_INSTRUCTION_NODE
	COMMENT_NODE
	DOCUMENT_NODE
	DOCUMENT_TYPE_NODE
	DOCUMENT_FRAGMENT_NODE
	NOTATION_NODE
	);

sub ELEMENT_NODE                () {1;}
sub ATTRIBUTE_NODE              () {2;}
sub TEXT_NODE                   () {3;}
sub CDATA_SECTION_NODE          () {4;}
sub ENTITY_REFERENCE_NODE       () {5;}
sub ENTITY_NODE                 () {6;}
sub PROCESSING_INSTRUCTION_NODE () {7;}
sub COMMENT_NODE                () {8;}
sub DOCUMENT_NODE               () {9;}
sub DOCUMENT_TYPE_NODE          () {10;}
sub DOCUMENT_FRAGMENT_NODE      () {11;}
sub NOTATION_NODE               () {12;}

our( $__nodeName,
	$__nodeValue,
	$__nodeType,
	$__parentNode,
	$__childNodes,
	$__attributes,
	$__canHaveChild
	) = ( 0, 1, 2, 3, 4, 5, 6 );

sub new {
	my ( $class, $name ) = @_;
	my $self = bless [], $class;

	$self->[$__nodeName] = $name;
	$self->[$__nodeValue] = '';
	$self->[$__nodeType] = 0;
	$self->[$__parentNode] = '';
	$self->[$__childNodes] = new XML::ESDOM_1_1::NodeList;

	$self;
}

sub nodeName {
	my $self = shift;
	$self->[$__nodeName];
}

sub nodeValue : lvalue {
	my $self = shift;
	$self->[$__nodeValue];
}

sub nodeType {
	my $self = shift;
	$self->[$__nodeType];
}

sub parentNode {
	my $self = shift;
	$self->[$__parentNode];
}

sub childNodes {
	my $self = shift;
	$self->[$__childNodes];
}

sub firstChild {
	my $self = shift;
	$self->childNodes->item(0);
}

sub lastChild {
	my $self = shift;
	$self->childNodes->item($self->childNodes->length-1);
}

sub previousSibling {
	my $self = shift;
	return unless( $self->parentNode );
	my $previous;
	for( my $i = 0;
		$i < $self->parentNode->childNodes->length;
		 $previous = $self->parentNode->childNodes->item($i), $i++ ) {
		if( $self->parentNode->childNodes->item($i) eq $self ) {
			last;
		}
	}
	$previous;
}

sub nextSibling {
	my $self = shift;
	return unless( $self->parentNode );
	my $next;
	for( my $i = $self->parentNode->childNodes->length -1;
		$i >= 0;
		 $next = $self->parentNode->childNodes->item($i), $i-- ) {
		if( $self->parentNode->childNodes->item($i) eq $self ) {
			last;
		}
	}
	$next;
}

sub attributes {
	my $self = shift;
	$self->[$__attributes];
}

sub insertBefore {
	my $self = shift;
	my ( $newChild, $refChild ) = @_;
	return if( !$newChild );
	return if( $refChild->parentNode ne $self );
	return if( $newChild->parentNode );
	return if( $newChild->_isAncestorOf( $self ) );
	return unless( $self->_canHaveChild );
	unless( $refChild ) {
		return $self->appendChild( $newChild );
	}
	return unless( $self->childNodes->_insertNode( $newChild, $refChild ) );
	$newChild->[$__parentNode] = $self;
	$newChild;
}

sub replaceChild {
	my $self = shift;
	my ( $newChild, $oldChild ) = @_;
	return if( !$newChild || !$oldChild );
	return if( $oldChild->parentNode ne $self );
	return if( $newChild->_isAncestorOf( $self ) );
	return unless( $self->_canHaveChild );
	if( $newChild->parentNode ) {
		$newChild->parentNode->removeChild( $newChild );
	}
	return unless( $self->childNodes->_replaceNode( $newChild, $oldChild ) );
	$newChild->[$__parentNode] = $self;
	$oldChild->[$__parentNode] = '';
	$oldChild;
}

sub removeChild {
	my $self = shift;
	my $oldChild = shift;
	return if( !$oldChild || $oldChild->parentNode ne $self );
	return unless( $self->childNodes->_removeNode( $oldChild ) );
	$oldChild->[$__parentNode] = '';
	$oldChild;
}

sub appendChild {
	my $self = shift;
	my $childNode = shift;
	return if( !$childNode );
	return if( $childNode->_isAncestorOf( $self ) );
	return unless( $self->_canHaveChild );
	if( $childNode->parentNode ) {
		$childNode->parentNode->removeChild( $childNode );
	}
	return unless( $self->childNodes->_addNode( $childNode ) );
	$childNode->[$__parentNode] = $self;
	$childNode;
}

sub hasChildNodes {
	my $self = shift;
	if( $self->childNodes && $self->childNodes->length ) {
		return 1;
	}
	return 0;
}

sub cloneNode {
	my $self = shift;
	my $deep = shift;
	my $clone;

	$clone = new XML::ESDOM_1_1::Node( $self->nodeName );
	$clone->nodeValue = $self->nodeValue;
	if( $self->attributes ) {
		for( my $i = 0; $i < $self->attributes->length; $i++ ) {
			$clone->attributes->setNamedItem( $self->attributes->item($i)->cloneNode( 0 ) );
		}
	}
	if( $deep ) {
		for( my $i = 0; $i < $self->childNodes->length; $i++ ) {
			$clone->appendChild( $self->childNodes->item($i)->cloneNode( 1 ) );
		}
	}
	$clone;
}

#############################
sub _canHaveChild : lvalue {
	$_[0]->[$__canHaveChild];
}

sub _isAncestorOf {
	my $self = shift;
	my $testNode = shift;

	while( my $parent = $testNode->parentNode ) {
		if( $parent eq $self ) {
			return 1;
		}
		$testNode = $parent;
	}
	0;
}


1;

